// @ts-check
import { init } from "@nulo/apkit";
import { execFile as _execFile } from "node:child_process";
import {
  chmod,
  copyFile,
  lstat,
  mkdir,
  mkdtemp,
  readFile,
  readdir,
  rm,
  writeFile,
} from "node:fs/promises";
import { tmpdir } from "node:os";
import { join } from "node:path";
import { promisify } from "node:util";
const execFile = promisify(_execFile);

const root = await mkdtemp(join(tmpdir(), "fireactions-rootfs."));
const alpine = await init(root);
await alpine.install(["util-linux", "dhcpcd", "eudev"]);

const initPath = await buildInit();
await copyFile(initPath, r("/sbin/fireactions-init"));

await writeSbin(
  r("/sbin/init"),
  `#!/bin/sh
/etc/init-files/00-pseudofs.sh
/etc/init-files/05-misc.sh

#exec /sbin/fireactions-init
/sbin/fireactions-init || sh
`
);

await mkdir(r("/etc/init-files/"));
await writeSbin(
  r("/etc/init-files/00-pseudofs.sh"),
  `#!/bin/sh
mountpoint -q /proc || mount -o nosuid,noexec,nodev -t proc proc /proc
mountpoint -q /sys || mount -o nosuid,noexec,nodev -t sysfs sys /sys
mountpoint -q /run || mount -o mode=0755,nosuid,nodev -t tmpfs run /run
mountpoint -q /dev || mount -o mode=0755,nosuid -t devtmpfs dev /dev
mkdir -p -m0755 /run/runit /run/lvm /run/user /run/lock /run/log /dev/pts /dev/shm
mountpoint -q /dev/pts || mount -o mode=0620,gid=5,nosuid,noexec -n -t devpts devpts /dev/pts
mountpoint -q /dev/shm || mount -o mode=1777,nosuid,nodev -n -t tmpfs shm /dev/shm
mountpoint -q /sys/kernel/security || mount -n -t securityfs securityfs /sys/kernel/security
`
);

await writeSbin(
  r("/etc/init-files/05-misc.sh"),
  `#!/bin/sh
  install -m0664 -o root -g utmp /dev/null /run/utmp
#halt -B  # for wtmp

ip link set up dev lo
hostname -F /etc/hostname
`
);

// await mkdirp(r("/etc/dropbear"));
// for (const t of ["rsa", "dss", "ed25519", "ecdsa"]) {
//   await execFile("dropbearkey", [
//     "-t",
//     t,
//     "-f",
//     r(`/etc/dropbear/dropbear_${t}_host_key`),
//   ]);
// }

// await execFile("sudo", ["mkdir", "-p", r("/root/.ssh")]);
// await writeFile(
//   "authorized_keys",
//   "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIEhrcTPMrhrOdwpgFkRFjTt8rdxt3gg16LJvBCZENRWa user@personal"
// );
// await execFile("sudo", [
//   "mv",
//   "authorized_keys",
//   r("/root/.ssh/authorized_keys"),
// ]);

// TODO: hacer squashfs
const ext4 = "rootfs.ext4";
await rm(ext4);
await execFile("fallocate", ["--length", "1G", ext4]);
await execFile("mkfs.ext4", [ext4]);

const mountpoint = await mkdtemp(join(tmpdir(), "tmp.fireactions-mountpoint."));
try {
  await mkdirp(mountpoint);
  await execFile("sudo", ["mount", ext4, mountpoint]);
  try {
    for (const dir of await readdir(root)) {
      await execFile("sudo", ["cp", "-r", r(dir), mountpoint]);
    }
  } finally {
    await execFile("sudo", ["umount", mountpoint]);
  }
} finally {
  await rm(mountpoint, { recursive: true });
}

// # sudo rm -rf "$dir"

// helpers
/**
 * @param {import("node:fs").PathLike} path
 * @param {string} content
 * @returns {Promise<void>}
 */
async function append(path, content) {
  const file = await readFile(path, "utf-8");
  await writeFile(path, file + "\n" + content);
}
/**
 * @param {import("node:fs").PathLike} path
 * @returns {Promise<void>}
 */
async function mkdirp(path) {
  await mkdir(path, { recursive: true });
}
/**
 * @param {string} path
 */
function r(path) {
  return join(root, path);
}

/**
 * @param {import("fs").PathLike} path
 * @param {string} content
 */
async function writeSbin(path, content) {
  try {
    await lstat(path);
    await rm(path, { recursive: true });
  } catch {}
  await writeFile(path, content);
  await chmod(path, 0o500);
}

async function buildInit() {
  const srcPath = "./rootfs/init";
  await execFile("podman", [
    ...["run", "--rm", "-it"],
    ...["-v", `${srcPath}:/home/rust/src:Z`],
    "docker.io/messense/rust-musl-cross:x86_64-musl",
    ...["cargo", "build", "--release"],
  ]);
  return join(srcPath, "target/x86_64-unknown-linux-musl/release/init");
}
